@@ -74,8 +74,8 @@ class Blockchain(object):
         self.lock = threading.RLock()
         
         # This is a dict of lists of pending blocks by the hash we need to
-        # verify them.
-        self.pending_on = collections.defaultdict(list)
+        # verify them, and then by their own hash (to detect duplicates)
+        self.pending_on = collections.defaultdict(dict)
         
         # This is a dict of callbacks by pending block hash
         self.pending_callbacks = {}
@@ -90,6 +90,12 @@ class Blockchain(object):
         if self.fork_data.has_key("state"):
             # Load the saved State.
             self.state = self.fork_data["state"]
+            
+        # We keep a cache of states after blocks, by block hash
+        self.state_cache = collections.OrderedDict()
+        
+        # How big should the cache be?
+        self.state_cache_limit = 10
 
     def has_block(self, block_hash):
         """
@@ -511,6 +517,9 @@ class Blockchain(object):
         If the block is eventually verifiable, call the callback with the hash
         and True if it's good, or the hash and False if it's bad.
         
+        If the same block is queued multiple times, and it isn't immediately
+        verifiable, only the last callback for the block will be called.
+        
         """
         
         # This holds all the callbacks we need to call, so we can call them
@@ -548,8 +557,12 @@ class Blockchain(object):
                         to_call.append((callback, (next_block.block_hash(),
                             False)))
             else:
-                # Add this block to the pending blocks
-                self.pending_on[next_block.previous_hash].append(next_block)
+                # Add this block to the pending blocks for its parent. If it's
+                # already there, we just replace it
+                self.pending_on[next_block.previous_hash][
+                    next_block.block_hash()] = next_block
+                # Remember the callback for this block, which chould be called
+                # when it is verified.
                 self.pending_callbacks[next_block.block_hash()] = callback
                 
                 print "Block height {}, hash {} pending on parent {}".format(
@@ -563,8 +576,7 @@ class Blockchain(object):
                 
                 # Pop off a hash to check
                 hash_added = added_hashes.pop()
-                
-                # Get the list of waiters
+                # Get the dict of waiters by waiter hash
                 waiters = self.pending_on[hash_added]
                 
                 # Remove it
@@ -573,15 +585,14 @@ class Blockchain(object):
                 print "{} waiters were waiting on {}".format(len(waiters), 
                     pybc.util.bytes2string(hash_added))
                     
-                for waiter in waiters:
+                for waiter_hash, waiter in waiters.iteritems():
                     # We ought to be able to verify and add each waiter.
                     
                     # Get the callback
-                    waiter_callback = self.pending_callbacks[
-                        waiter.block_hash()]
+                    waiter_callback = self.pending_callbacks[waiter_hash]
                     
                     # Remove it from the collection of remaining callbacks
-                    del self.pending_callbacks[waiter.block_hash()]
+                    del self.pending_callbacks[waiter_hash]
                 
                     if self.can_verify_block(waiter):
                         # We can verify the block right now (ought to always be
@@ -593,27 +604,27 @@ class Blockchain(object):
                             if waiter_callback is not None:
                                 # Call the callback later
                                 to_call.append((waiter_callback, 
-                                    (waiter.block_hash(), True)))
+                                    (waiter_hash, True)))
                             
                             print "Queued block height {} verified: {}".format(
                                 waiter.height,
-                                pybc.util.bytes2string(waiter.block_hash()))
+                                pybc.util.bytes2string(waiter_hash))
                             
                             # Record that we added the pending block, so things
                             # pending on it can now be added
-                            added_hashes.append(waiter.block_hash())
+                            added_hashes.append(waiter_hash)
                         else:
                             # TODO: throw out blocks waiting on invalid blocks.
                             # If we have any of those, there's probablt a hard
                             # fork.
                             
                             print "Queued block invalid: {}".format(
-                                pybc.util.bytes2string(waiter.block_hash()))
+                                pybc.util.bytes2string(waiter_hash))
                             
                             if waiter_callback is not None:
                                 # Call the callback later
-                                to_call.append((waiter_callback, 
-                                    (waiter.block_hash(), False)))
+                                to_call.append((waiter_callback, (waiter_hash, 
+                                    False)))
                     else:
                         # This should never happen
                         print("Couldn't verify a waiting block when its parent "
@@ -626,32 +637,100 @@ class Blockchain(object):
             # another thread has and that thread is waiting on this thread.
             callback(*args)
     
+    def cache_state(self, block_hash, state):
+        """
+        Put this State object in the state cache as the state after the block
+        with the given hash. Make it the most recently used state, and eject
+        older states if needed to make this state fit.
+        
+        """
+    
+        if self.state_cache.has_key(block_hash):
+            # Take it out so we can move it to the end.
+            del self.state_cache[block_hash]
+        elif len(self.state_cache) == self.state_cache_limit:
+            # Pop the oldes thing because it's full
+            self.state_cache.popitem()
+        
+        # Add the entry
+        self.state_cache[block_hash] = state
+            
+    
     def state_after(self, block):
         """
         Given that our State is currently up to date with our current
         highest_block, return a State that is in effect after the given block.
         
-        block may not be None.
+        block may be None.
         
         Walks the blockchain from the tip of the highest fork to the given
         block, which must be in the block store.
         
+        Because that's probably what's responsible for poor performance when
+        trying to fork, we keep an LRU cache of states.
+        
+        TODO: can we do real pathfinding?
+        
         """
         
         with self.lock:
         
+            # What's the hash of the block?
+            if block is not None:
+                block_hash = block.block_hash()
+            else:
+                block_hash = "\0" * 64
+            
+            if self.state_cache.has_key(block_hash):
+                print "Getting cached state"
+                # We already have the answer. Get it from the cache.
+                state = self.state_cache[block_hash]
+                
+                # Mark the state used 
+                self.cache_state(block_hash, state)
+                
+                # Return it
+                return state
+        
             if block == self.highest_block:
+                print "Using current state."
                 # We already have the state after this
+                self.cache_state(block_hash, self.state)
                 return self.state
         
             if self.highest_block is None:
+                print "Making genisis state"
                 # Easy case: this must be a genesis block.
-                # Return the state updated with the new block
-                return self.state.step_forwards(block)
+                # Use the current (empty) state updated with the new block
+                state = self.state.step_forwards(block)
+                self.cache_state(block_hash, state)
+                return state
+                
+            if (block is not None and 
+                self.state_cache.has_key(block.previous_hash)):
+                print "Using parent state"
+                # Special case: we have the state after the block before this
+                # block, so we can just add this block on top.
+                # TODO: real states and pathfinding
+                state = self.state_cache[block.previous_hash]
+                
+                # Mark the state used 
+                self.cache_state(block.previous_hash, state)
+                
+                # Advance the State
+                state = state.step_forwards(block)
+                
+                # Cache and return it
+                self.cache_state(block_hash, state)
+                return state
             
+            print "Walking blockchain for state"
             # If we get here, we know we have a highest block, and that we need
             # to walk from there to the block we're being asked about.
             
+            # How many blocks did we walk?
+            blocks_walked = 0
+            
             # Find the common ancestor of this block and the tip, where we have
             # a known State.
             ancestor = self.common_ancestor(block, self.highest_block)
@@ -685,6 +764,7 @@ class Blockchain(object):
                 
                 # Step back a block
                 position = self.blockstore[position].previous_hash
+                blocks_walked += 1
                 
             # Now we've reverted to the post-common-ancestor state. TODO: Can't
             # we just get a known state here? Or can we save a state per tip?
@@ -715,6 +795,7 @@ class Blockchain(object):
                 
                 # Step back a block
                 position = self.blockstore[position].previous_hash
+                blocks_walked += 1
                 
             # Flip the blocks that need to be applied into chronological order.
             blocks_to_apply.reverse()
@@ -723,7 +804,10 @@ class Blockchain(object):
                 # Apply the block to the state
                 state = state.step_forwards(block_to_apply)
             
+            print "Walked {} blocks".format(blocks_walked)
+            
             # We've now updated to the state for after the given block.
+            self.cache_state(block_hash, state)
             return state
         
         
