@@ -22,7 +22,7 @@ class Peer(object):
     
     def __init__(self, network, version, blockchain, peer_file = ":memory:", 
         external_address=None, port=8007, optimal_connections=10, 
-        tick_period=60, peer_timeout = 60 * 60 * 12):
+        tick_period=60, peer_timeout = 60 * 60 * 48):
         """
         Make a new Peer in the given network (identified by a string), with the
         given version integer, that keeps its blocks in the given Blockchain.
@@ -247,29 +247,35 @@ class Peer(object):
                 # We don't have enough outgoing connections, but we do know some
                 # peers.
                 
-                # Find a peer we aren't connected to and connect to them
-                host = random.sample(self.known_peers, 1)[0]
+                for _ in xrange(self.optimal_connections - 
+                    len(self.outgoing_hosts)):
                 
-                # Try at most 100 times to find a host we aren't connected to
-                attempt = 1
-                while host in self.outgoing_hosts and attempt < 100:
-                    # Try a new host
+                    # For each connection we want but don't have
+                
+                    # Find a peer we aren't connected to and connect to them
                     host = random.sample(self.known_peers, 1)[0]
                     
-                    # Increment attempt
-                    attempt += 1
+                    # Try at most 100 times to find a host we aren't connected
+                    # to
+                    attempt = 1
+                    while host in self.outgoing_hosts and attempt < 100:
+                        # Try a new host
+                        host = random.sample(self.known_peers, 1)[0]
+                        
+                        # Increment attempt
+                        attempt += 1
+                        
+                    # TODO: This always makes two attempts at least
+                        
+                    if attempt < 100:
+                        # We found one!
+                        # Connect to it.
+                        
+                        # Get the port (and discard the last heard from time)
+                        port, _ = self.known_peers[host]
                     
-                # TODO: This always makes two attempts at least
-                    
-                if attempt < 100:
-                    # We found one!
-                    # Connect to it.
-                    
-                    # Get the port (and discard the last heard from time)
-                    port, _ = self.known_peers[host]
-                
-                    # Connect to it.
-                    self.connect(host, port)
+                        # Connect to it.
+                        self.connect(host, port)
                     
             # Throw out peers that are too old. First compile a list of their
             # hostnames.
@@ -387,14 +393,17 @@ class Peer(object):
     
     def send_block(self, block):
         """
-        Given a new block, will add it to our blockchain and send it on its way
-        to all peers, if appropriate.
+        Given a (possibly new) block, will add it to our blockchain and send it
+        on its way to all peers, if appropriate.
         
         """
         
-        # Queue the block, and, if valid, announce it
-        self.blockchain.queue_block(block, 
-            callback=self.was_block_valid)
+        
+        if not self.blockchain.has_block(block.block_hash()):
+            # It's new.
+            # Queue the block, and, if valid, announce it
+            self.blockchain.queue_block(block, 
+                callback=self.was_block_valid)
             
     def send_transaction(self, transaction):
         """ 
